"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _path = _interopRequireDefault(require("path"));

var _lodash = _interopRequireDefault(require("lodash"));

var _postgres = _interopRequireDefault(require("postgres"));

var _commonsApi = require("@verdaccio/commons-api");

var _pgFs = _interopRequireWildcard(require("./pg-fs"));

var _levelAdapter = require("./level-adapter");

var _setConfigValue = _interopRequireDefault(require("./setConfigValue"));

function _getRequireWildcardCache() { if (typeof WeakMap !== "function") return null; var cache = new WeakMap(); _getRequireWildcardCache = function () { return cache; }; return cache; }

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function ownKeys(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

const TABLE_NAME = 'verdaccio';
const TOKEN_TABLE_NAME = 'tokens';

class PGDatabase {
  constructor(config, options) {
    _defineProperty(this, "logger", void 0);

    _defineProperty(this, "config", void 0);

    _defineProperty(this, "locked", void 0);

    _defineProperty(this, "tokenDb", void 0);

    _defineProperty(this, "sql", void 0);

    _defineProperty(this, "data", void 0);

    this.locked = false;
    this.logger = options.logger;

    if (!config) {
      throw new Error('pg storage missing config. Add `store.pg-storage` to your config file');
    }

    this.config = Object.assign(config, config.store['pg-storage']);
    this.config.url = (0, _setConfigValue.default)(this.config.url);
    this.config.host = (0, _setConfigValue.default)(this.config.host);
    this.config.port = (0, _setConfigValue.default)(this.config.port);
    this.config.path = (0, _setConfigValue.default)(this.config.path);
    this.config.database = (0, _setConfigValue.default)(this.config.database);
    this.config.username = (0, _setConfigValue.default)(this.config.username);
    this.config.password = (0, _setConfigValue.default)(this.config.password);
    this.config.ssl = (0, _setConfigValue.default)(this.config.ssl);
    this.config.max = (0, _setConfigValue.default)(this.config.max);
    this.config.idle_timeout = (0, _setConfigValue.default)(this.config.idle_timeout);
    this.config.connect_timeout = (0, _setConfigValue.default)(this.config.connect_timeout);
    this.logger.debug({
      config: JSON.stringify(this.config, null, 4)
    }, 'pg: configuration: @{config}');

    if (this.config.url) {
      this.sql = (0, _postgres.default)(this.config.url, {
        host: this.config.host,
        port: this.config.port ? +this.config.port : undefined,
        path: this.config.path,
        database: this.config.database,
        username: this.config.username,
        password: this.config.password,
        ssl: this.config.ssl === 'false' || this.config.ssl === 'true' ? this.config.ssl === 'true' : undefined,
        max: this.config.max ? +this.config.max : undefined,
        idle_timeout: this.config.idle_timeout ? +this.config.idle_timeout : undefined,
        connect_timeout: this.config.connect_timeout ? +this.config.connect_timeout : undefined
      });
    } else {
      this.sql = (0, _postgres.default)({
        host: this.config.host,
        port: this.config.port ? +this.config.port : undefined,
        path: this.config.path,
        database: this.config.database,
        username: this.config.username,
        password: this.config.password,
        ssl: this.config.ssl === 'false' || this.config.ssl === 'true' ? this.config.ssl === 'true' : undefined,
        max: this.config.max ? +this.config.max : undefined,
        idle_timeout: this.config.idle_timeout ? +this.config.idle_timeout : undefined,
        connect_timeout: this.config.connect_timeout ? +this.config.connect_timeout : undefined
      });
    }

    this.data = this._fetchLocalPackages();
    this.tokenDb = (0, _levelAdapter.level)(this.sql, TOKEN_TABLE_NAME);
    this.logger.trace({
      config: this.config
    }, '[pg-store]: configuration: @{config}');

    this._sync();
  }

  async getSecret() {
    return (await this.data).secret;
  }

  setSecret(secret) {
    this.data = this.data.then(data => _objectSpread({}, data, {
      secret
    }));
    return this._sync();
  }

  async add(name, cb) {
    if (!(await this.data).list.includes(name)) {
      this.data = this.data.then(data => _objectSpread({}, data, {
        list: data.list.concat(name)
      }));
      this.logger.debug({
        name
      }, '[pg-storage]: the private package @{name} has been added');
      cb(await this._sync());
    } else {
      cb(null);
    }
  }

  async search(onPackage, onEnd) {
    try {
      await this.sql`
        SELECT path, content, updated_at FROM ${this.sql(_pgFs.TABLE_NAME)} WHERE path LIKE '%/package.json'
      `.stream(({
        path,
        content,
        updated_at: LastModified
      }) => {
        const pkg = JSON.parse(content.toString('utf8'));
        return onPackage({
          name: pkg.name,
          path: path,
          time: LastModified.getTime()
        });
      });
      onEnd();
    } catch (err) {
      onEnd(err);
    }
  }

  remove(name, cb) {
    this.get(async (err, list) => {
      if (err) {
        cb((0, _commonsApi.getInternalError)('error remove private package'));
        this.logger.error({
          err
        }, '[pg-storage/remove]: remove the private package has failed @{err}');
      }

      this.data = this.data.then(data => _objectSpread({}, data, {
        list: list.filter(pkgName => pkgName !== name)
      }));
      cb(await this._sync());
    });
  }

  async get(cb) {
    const list = (await this.data).list;
    const totalItems = (await this.data).list.length;
    cb(null, list);
    this.logger.trace({
      totalItems
    }, '[pg-storage/get]: full list of packages (@{totalItems}) has been fetched');
  }

  getPackageStorage(packageName) {
    const packageAccess = this.config.getMatchedPackagesSpec(packageName);

    const packagePath = this._getLocalStoragePath(packageAccess ? packageAccess.storage : undefined);

    this.logger.trace({
      packagePath
    }, '[pg-storage/getPackageStorage]: storage selected: @{packagePath}');

    if (_lodash.default.isString(packagePath) === false) {
      this.logger.debug({
        name: packageName
      }, 'this package has no storage defined: @{name}');
      return;
    }

    const packageStoragePath = _path.default.join(_path.default.resolve(_path.default.dirname(this.config.self_path || ''), packagePath), packageName);

    this.logger.trace({
      packageStoragePath
    }, '[pg-storage/getPackageStorage]: storage path: @{packageStoragePath}');
    return new _pgFs.default(this.sql, packageStoragePath, this.logger);
  }

  clean() {
    this._sync();
  }

  saveToken(token) {
    const key = this._getTokenKey(token);

    const db = this.tokenDb;
    return new Promise((resolve, reject) => {
      db.put(key, token, err => {
        if (err) {
          reject(err);
          return;
        }

        resolve();
      });
    });
  }

  deleteToken(user, tokenKey) {
    const key = this._compoundTokenKey(user, tokenKey);

    const db = this.tokenDb;
    return new Promise((resolve, reject) => {
      db.del(key, err => {
        if (err) {
          reject(err);
          return;
        }

        resolve();
      });
    });
  }

  readTokens(filter) {
    return new Promise((resolve, reject) => {
      const tokens = [];
      const key = filter.user + ':';
      const db = this.tokenDb;
      const stream = db.createReadStream({
        gte: key,
        lte: String.fromCharCode(key.charCodeAt(0) + 1)
      });
      stream.on('data', data => {
        tokens.push(data.value);
      });
      stream.once('end', () => resolve(tokens));
      stream.once('error', err => reject(err));
    });
  }

  async _fetchLocalPackages() {
    const list = [];
    const emptyDatabase = {
      list,
      secret: ''
    };

    try {
      const [{
        value: value
      }] = await this.sql`SELECT value FROM ${this.sql(TABLE_NAME)}`;
      return value;
    } catch (err) {
      // Only recreate if table not found to prevent data loss undefined_table
      if (err.code !== '42P01') {
        this.locked = true;
        this.logger.error('Failed to read package database table, please check the error printed below:\n', `Table Name: ${TABLE_NAME}\n\n ${err.message}`);
      }

      return emptyDatabase;
    }
  }

  async _sync() {
    this.logger.debug('[pg-storage/_sync]: init sync database');

    if (this.locked) {
      this.logger.error('Database is locked, please check error message printed during startup to prevent data loss.');
      return new Error('Verdaccio database is locked, please contact your administrator to checkout logs during verdaccio startup.');
    }

    try {
      await this.sql`CREATE TABLE IF NOT EXISTS ${this.sql(TABLE_NAME)} (
        key bool PRIMARY KEY DEFAULT TRUE,
        value jsonb,
        created_at timestamp not null default current_timestamp,
        updated_at timestamp not null default current_timestamp
        CONSTRAINT local_packages_unique CHECK (key)
      )`;
      this.logger.debug({
        tableName: TABLE_NAME
      }, '[pg-storage/_sync]: table @{tableName} created succeed');
    } catch (err) {
      this.logger.debug({
        err
      }, '[pg-storage/_sync/create-table-if-not-exists]: sync failed @{err}');
      return null;
    }

    try {
      await this.sql`
        INSERT INTO ${this.sql(TABLE_NAME)} (key, value, updated_at) VALUES (TRUE, ${this.sql.json(await this.data)}, NOW())
        ON CONFLICT (key) DO UPDATE SET value = ${this.sql.json(await this.data)}
      `;
      this.logger.debug('[pg-store/_sync/update_statement]: sync write succeed');
      return null;
    } catch (err) {
      this.logger.debug({
        err
      }, '[pg-store/_sync/update/statement]: sync failed$ @{err}');
      return err;
    }
  }

  _getLocalStoragePath(storage) {
    const globalConfigStorage = this.config ? this.config.storage : undefined;

    if (_lodash.default.isNil(globalConfigStorage)) {
      throw new Error('global storage is required for this plugin');
    } else {
      if (_lodash.default.isNil(storage) === false && _lodash.default.isString(storage)) {
        return _path.default.join(globalConfigStorage, storage);
      }

      return globalConfigStorage;
    }
  }

  _getTokenKey(token) {
    const {
      user,
      key
    } = token;
    return this._compoundTokenKey(user, key);
  }

  _compoundTokenKey(user, key) {
    return `${user}:${key}`;
  }

}

var _default = PGDatabase;
exports.default = _default;