/// <reference types="node" />
import { ILocalPackageManager, Logger, Callback, Package, IUploadTarball } from '@verdaccio/types';
import { VerdaccioError } from '@verdaccio/commons-api/lib';
import postgres from 'postgres';
import { ReadTarball } from '@verdaccio/streams';
export declare const TABLE_NAME = "files";
export declare const PKG_FILE_NAME = "package.json";
export declare const ERROR_NO_SUCH_FILE: VerdaccioError;
export declare const ERROR_FILE_EXIST: VerdaccioError;
export declare type IPGPackageManager = ILocalPackageManager & {
    prefix: string;
};
export default class PGPackageManager implements IPGPackageManager {
    prefix: string;
    logger: Logger;
    private sql;
    private table;
    constructor(sql: postgres.Sql<never>, prefix: string, logger: Logger);
    updatePackage(name: string, updateHandler: Callback, onWrite: Callback, transformPackage: Function, onEnd: Callback): Promise<void>;
    deletePackage(packageName: string, callback: (err: Error | null) => void): Promise<void>;
    removePackage(callback: (err: NodeJS.ErrnoException | null) => void): Promise<void>;
    createPackage(name: string, value: Package, callback: Callback): void;
    savePackage(name: string, value: Package, callback: Callback): void;
    readPackage(name: string, callback: Callback): Promise<void>;
    writeTarball(name: string): IUploadTarball;
    readTarball(name: string): ReadTarball;
    private _createFile;
    private _readStorageFile;
    private _convertToString;
    private _getStorage;
    private _writeFile;
    private _deleteFile;
    private _deletePrefix;
    private _createTable;
}
